#include "config.h"
#include "uart.h"
#include "w25qxx.h"

BYTE xdata pRxBuffer[256];          //UARTݽջ
BYTE xdata pFlashBuffer[256];       //Flashݻ
BYTE bReadPtr;                      //ָ
BYTE bWritePtr;                     //дָ
BOOL bTxBusy;                       //UARTڷæ־

void UART_Init()
{
    SCON = 0x50;
    T2L = BAUD;
    T2H = BAUD >> 8;
    AUXR |= 0x15;
    ES = 1;

    UART_Flush();
    bTxBusy = FALSE;
}

void UART_Isr() interrupt UART1_VECTOR
{
    if (TI)
    {
        TI = 0;
        bTxBusy = FALSE;
    }
    
    if (RI)
    {
        RI = 0;
        pRxBuffer[bWritePtr++] = SBUF;
    }
}

void UART_Flush()
{
    bReadPtr = 0;
    bWritePtr = 0;
}

void UART_SendData(BYTE *dat, int size)
{
    while (size--)
        putchar(*dat++);
}

char putchar(char dat)
{
    while (bTxBusy);
    SBUF = dat;
    bTxBusy = TRUE;
    
    return dat;
}

char _getkey()
{
    while (bReadPtr == bWritePtr);
    
    return pRxBuffer[bReadPtr++];
}

void UART_Process()
{
#define CMD_READ_ID             0xf0
#define CMD_READ_STATUS         0xf1
#define CMD_READ_DATA           0xf2
#define CMD_PROGRAM_DATA        0xf3
#define CMD_ERASE_CHIP          0xf4
#define CMD_ERASE_4K            0xf5
#define CMD_ERASE_32K           0xf6
#define CMD_ERASE_64K           0xf7

#define STG_IDLE                0x00
#define STG_CMD_HEAD1           0x01
#define STG_CMD_HEAD2           0x02
#define STG_CMD_HEAD3           0x03
#define STG_ERASE_ADDR0         0x10
#define STG_ERASE_ADDR1         0x11
#define STG_ERASE_ADDR2         0x12
#define STG_ERASE_ADDR3         0x13
#define STG_ERASE_DONE          0x14
#define STG_READ_ADDR0          0x20
#define STG_READ_ADDR1          0x21
#define STG_READ_ADDR2          0x22
#define STG_READ_ADDR3          0x23
#define STG_READ_LEN0           0x24
#define STG_READ_LEN1           0x25
#define STG_READ_LEN2           0x26
#define STG_READ_LEN3           0x27
#define STG_READ_DONE           0x28
#define STG_PROGRAM_ADDR0       0x30
#define STG_PROGRAM_ADDR1       0x31
#define STG_PROGRAM_ADDR2       0x32
#define STG_PROGRAM_ADDR3       0x33
#define STG_PROGRAM_LEN0        0x34
#define STG_PROGRAM_LEN1        0x35
#define STG_PROGRAM_LEN2        0x36
#define STG_PROGRAM_LEN3        0x37
#define STG_PROGRAM_DATA        0x38

#define ISADDR32()              ((BYTE)(addr >> 24))

    static BYTE stage;
    static BYTE cmd;
    static DWORD addr;
    static DWORD len;
    static WORD index;
    BYTE dat;
    BYTE cnt;
    
    while (bReadPtr != bWritePtr)
    {
        dat = _getkey();
        switch (stage)
        {
        default:
        case STG_IDLE:
L_CheckHead:
            if (dat == 'A')
                stage = STG_CMD_HEAD1;
            else
                stage = STG_IDLE;
            break;
        case STG_CMD_HEAD1:
            if (dat == 'I')
                stage = STG_CMD_HEAD2;
            else
                goto L_CheckHead;
            break;
        case STG_CMD_HEAD2:
            if (dat == 'C')
                stage = STG_CMD_HEAD3;
            else
                goto L_CheckHead;
            break;
        case STG_CMD_HEAD3:
            addr = 0;
            len = 0;
            index = 0;
            switch (cmd = dat)
            {
            case CMD_READ_ID:
                UART_SendData("AIS", 3);
                *(DWORD *)pFlashBuffer = W25Q_ReadJEDECID_9F();
                UART_SendData(&pFlashBuffer[1], 3);
                stage = STG_IDLE;
                break;
            case CMD_READ_STATUS:
                UART_SendData("AIS", 3);
                pFlashBuffer[0] = W25Q_ReadSR1_05();
                pFlashBuffer[1] = W25Q_ReadSR2_35();
                pFlashBuffer[2] = W25Q_ReadSR3_15();
                UART_SendData(pFlashBuffer, 3);
                stage = STG_IDLE;
                break;
            case CMD_READ_DATA:
                stage = STG_READ_ADDR0;
                break;
            case CMD_PROGRAM_DATA:
                stage = STG_PROGRAM_ADDR0;
                break;
            case CMD_ERASE_CHIP:
                W25Q_EraseChip_C7(FALSE);
                UART_SendData("AIS", 3);
                stage = STG_IDLE;
                break;
            case CMD_ERASE_4K:
            case CMD_ERASE_32K:
            case CMD_ERASE_64K:
                stage = STG_ERASE_ADDR0;
                break;
            default:
                goto L_CheckHead;
            }
            break;
        case STG_READ_ADDR0:
        case STG_READ_ADDR1:
        case STG_READ_ADDR2:
        case STG_READ_ADDR3:
        case STG_PROGRAM_ADDR0:
        case STG_PROGRAM_ADDR1:
        case STG_PROGRAM_ADDR2:
        case STG_PROGRAM_ADDR3:
        case STG_ERASE_ADDR0:
        case STG_ERASE_ADDR1:
        case STG_ERASE_ADDR2:
        case STG_ERASE_ADDR3:
            addr <<= 8;
            addr |= dat;
            stage++;
            if (stage == STG_ERASE_DONE)
            {
                switch (cmd)
                {
                case CMD_ERASE_4K:
                    if (ISADDR32())
                        W25Q_Erase4K_Addr32_21(addr, FALSE);
                    else
                        W25Q_Erase4K_20(addr, FALSE);
                    UART_SendData("AIS", 3);
                    stage = STG_IDLE;
                    break;
                case CMD_ERASE_64K:
                    if (ISADDR32())
                        W25Q_Erase64K_Addr32_DC(addr, FALSE);
                    else
                        W25Q_Erase64K_D8(addr, FALSE);
                    UART_SendData("AIS", 3);
                    stage = STG_IDLE;
                    break;
                default:
                    goto L_CheckHead;
                }
            }
            break;
        case STG_READ_LEN0:
        case STG_READ_LEN1:
        case STG_READ_LEN2:
        case STG_READ_LEN3:
        case STG_PROGRAM_LEN0:
        case STG_PROGRAM_LEN1:
        case STG_PROGRAM_LEN2:
        case STG_PROGRAM_LEN3:
            len <<= 8;
            len |= dat;
            stage++;
            if (stage == STG_READ_DONE)
            {
                UART_SendData("AIS", 3);
                while (len)
                {
                    cnt = min(len, 64);
                    if (ISADDR32())
                        W25Q_FastRead_Addr32_6C(addr, pFlashBuffer, cnt);
                    else
                        W25Q_FastRead_6B(addr, pFlashBuffer, cnt);

                    UART_SendData(pFlashBuffer, cnt);
                    
                    addr += cnt;
                    len -= cnt;
                }
                stage = STG_IDLE;
            }
            break;
        case STG_PROGRAM_DATA:
            pFlashBuffer[index++] = dat;
            if (index >= len)
            {
                if (ISADDR32())
                    W25Q_PageProgram_Addr32_12(addr, pFlashBuffer, (WORD)len);
                else
                    W25Q_PageProgram_02(addr, pFlashBuffer, (WORD)len);
                
                UART_SendData("AIS", 3);
                stage = STG_IDLE;
            }
            break;
        }
    }
}
